#!/usr/local/bin/perl -w

#----------------------------------------------------------------------
# Cristal:
#	Archana Bellamkonda
#   Rashmi V Kankaria
#   Ashutosh A Nagle
#   Paul J Gordon
# 
# 12/16/2002
# Final Project 
# CS 8761 - Natural Language Processing - Fall 2002 
#
# Format: sentiment.pl Dir
#
# sentiment.pl returns a positive or negetive sentiment assesment 
# for text files.  Dir is a directory that contains two sub-directories.
# One must be named pos, it contains the files with a positive semantic
# orientation.  The other sub-directory must be named neg, and it contains
# the files with a negative semantic orientation.  The program returns the
# percentage correct.  It also maintains a log file, trace.txt which keeps
# track of the name of each file assesed, what the assessment was ( positive
# or negative), and what the actual orientation of the file was.  When all the
# files have been processsed, it returns the percent correct, the precission,
# and the recall for all the files in the neg and pos subdirectories of Dir.
#
# Thanks to Peter Turney.  The paper below got us started with our algorithm.
#
# Peter Turney.  2002.  Thumbs up or thumbs down?  Semantic orientation applied
#   to unsupervised learning of semantic clasification fo reivews.  In Proc. of
#   the ALC.
#
#-------------------------------------------------------------------------------

use LDOCE::ldoc;
use WebReader::webSearch ;
use BigMac::QueryMac ;


my $precision = 0;
my $recall = 0;

my $temp = 0;



# Efficiency check
my $time = time ;

# Call the classifier, return the percent correct
my $correct = sentimentClassifier(shift (@ARGV),$precision,$recall)*100;
print "Percentage Correct = $correct\%\n";
print "Execution time = ", (time - $time)/60 , " minutes\n" ;

# getWebSentences takes two strings as input.  Each string begins with
# its semantic orientation.  and then a list of representative words.
# These words are used as input to the webSearch program (via the
# function makeSentenceList), and a list of sentences from the web, containing
# the representative words is returned.
# This function also checks to see if cache files exits.  If the do, sentences
# are read from the cache file, rather than from the web.
#
sub getWebSentences 
{
    my $goodWordList = shift ;
    my $badWordList = shift ;

	# remove orientation word and
	# split representative(seed) words into an array.
    my @goodList = split ( /:\s*/ , $goodWordList ) ;
    my @goodListNew = split ( /\s*,\s*/ , $goodList[1] ) ;
	
	my @badList = split ( /:\s*/ , $badWordList ) ;
    my @badListNew = split ( /\s*,\s*/ , $badList[1] ) ; 


	# if both cache files don't exist, call the webSearch module(via makeSentenceList)
    # and create the cache files.
	if( ! (-e "__goodWebSentencesFile") || ! (-e "__badWebSentencesFile") ) {
	@good=makeSentenceList("@goodListNew") ;
	
	open $goodListToFile, "> __goodWebSentencesFile" or die "Couldn't open __goodWebSentencesFile for writting: $!\n" ;
	system ("chmod 777 __goodWebSentencesFile");
	foreach  (@good) {print $goodListToFile "$_\n" ;}
	
	@bad=makeSentenceList("@badListNew") ;
	
	open $badListToFile, ">__badWebSentencesFile" or die "Couldn't open __badWebSentencesFile for writting: $!\n" ;
	system ("chmod 777 __badWebSentencesFile");
	foreach  (@bad) {print $badListToFile "$_\n" ;}
    }
    # files exist, so write sentences into arrays.
    else {
	open $goodListFromFile, "__goodWebSentencesFile" or die "Couldn't open  __goodWebSentencesFile for reading: $!\n" ;
	while(<$goodListFromFile>) { push @good, $_ ; }
	open $badListFromFile, "__badWebSentencesFile" or die "Couldn't open __badWebSentencesFile for reading: $!\n" ;
	while(<$badListFromFile>) { push @bad, $_ ; }
    }
}

#This function returns relevant words from the reveiw. This make use of
#the information from web to do this.
sub giveMeWordsIn {

    my $reviewfile = shift ;
    my $fh = ReviewFileinGiveMeWordsIn ;
    my $reviewString ;

    my $goodSentenceCount = scalar @good ;
    my $badSentenceCount = scalar @bad ;
    
    my $probGood = 0 ;
    my $probBad = 0 ;
    my $wordResult = 0 ;

    my @filteredPhrases ;
    my @phrases ;

    open $fh, "$reviewfile" or die "Couldn't open reviewfile for reading: $!\n" ;
    while (<$fh> ) { chomp ; $reviewString.=$_ ; }
    $_ = $reviewString ;
    #print $_ ;

    @phrases = /\b\w+?\b/g ;
    #foreach(@phrases) { print "*$_*\n" ; }

    foreach $phrase ( @phrases ) {
	$phrase = uc($phrase) ;
	#print "PHRASE: *$phrase*\n" ;
	$badPhrases = 0 ;
	$goodPhrases = 0 ;
	#print "Phrase: $phrase\n" ;
	foreach ( @good ) {
	    #print "Good Sentence: $_\n" ;
	    if( /$phrase/ ) { $goodPhrases++ ; }
	}
	foreach ( @bad ) {
	    #print "Bad Sentences: $_\n" ;
	    if ( /$phrase/ ) { $badPhrases++ ; }
	}
	
	$probGood = $goodPhrases / $goodSentenceCount ;
	$probBad = $badPhrases / $badSentenceCount ; 


	if ( $probGood == 0 ) { $probGood = .01 ; }
	if ( $probBad == 0 ) { $probBad = .01 ; }

	if( $probGood <= $probBad ) { $wordResult = ( $probBad - $probGood ) / $probGood  ; }
	else { $wordResult = ( $probGood - $probBad ) / $probBad  ; }
	
	#print "Result: $wordResult\n" ;

	if( $wordResult > .5 ) { push @filteredPhrases , $phrase ; }
    
    }
    #Remove words in stoplist
    if (open (STOP, "<stopList"))
    {
	my @myTempFL = ();	
	my $sl = <STOP>;
	close STOP;
	chomp($sl);
	my @SL = split (/\s*,\s*/,$sl);
	while (@filteredPhrases)
	{
	    my $phrase = pop(@filteredPhrases);
	    my $put = 1;
	    foreach my $stopWord (@SL)
	    {
		if ($stopWord eq $phrase)
		{
		    $put = 0;
		    last;
		}
	    }

	    if ($put)
	    {
		push (@myTempFL, $phrase);
	    }
	}
	return @myTempFL;
    }
	
    return @filteredPhrases ;
    
}


#This function processes a line in the seed file.
#It takes in the line, separates the word using other functions
#Gets their definitions that ar associated with the parts of speech
#adj or adv and finally extracts and returns the adjectives and adverbs
#from those definitions.

sub processSeedLine
{
    my $line = $_[0];
    my %purgedContextWords = ();

    if ($line !~ /:/)
    {
	return %purgedContextWords;
    }
    my $fileName = $`;
    $fileName =~ s/\s+$//;
    $fileName = "__".$fileName."ContextFile";
    my $fh = FILEinProcessSeedLine;
    my $genFile = 0;

    if (!(open $fh,"<$fileName"))
    {
	$genFile = 1;
    }
    else
    {
	my $firstLine = <$fh>;
	chomp($firstLine);
	if ($firstLine ne $line)
	{
	    $genFile = 1;
	    close $fh;
	}
    }
    
    if ($genFile)
    {
	my $context = createContext ($line);
	my @contextWords = split (/([\s\n\t,.!\?])+/,$context); 
	my %purgedContextWords = purge(@contextWords);
	    
	if (!(open $fh,">$fileName"))
	{
	    return %purgedContextWords;
	}
	system ("chmod 777 $fileName");
	chomp($line);
	print $fh "$line\n";
	foreach (sort keys %purgedContextWords)
	{
	    print $fh "$_\n";
	}
	close $fh;
    } else
    {
	while (my $line = <$fh>)
	{
	    chomp($line);
	    $purgedContextWords{$line} = 0;
	}
	close $fh;
    }
    return %purgedContextWords;
}


# sentimentClassifier controls the main flow of the program.  It calls the other
# subroutines that it will need to create the context, and to assign scores to
# to reviews.  It also calculates the precission and recall for all the text files
# that it assigns values for.
#
sub sentimentClassifier
{
    open TRACE, ">traceMoviePaul.txt"; # Detailed information about results
    system( "chmod 777 traceMoviePaul.txt" ) ;
    print TRACE "INTERPRETATION:\n\tIf the score of a review is greater than 0.5, it's a positive review. \n\tIf it is less than -0.5, it's a negative review.\n\tOtherwise it's neutral.\n\n";

    my $dir = $_[0]; # get directory name from command line
    my $dh = DIRinSentimentClassifier; #Directory handle in this function

#Process the first seed file line (good line)
    open SEED, "<seedFile";
    my $seedLine1 = <SEED>;
    chomp($seedLine1);
    my %purgedGoodContextWords = processSeedLine ($seedLine1); 

#Process the second seed file line (bad line)
    my $seedLine2 = <SEED>;
    chomp($seedLine2);
    my %purgedBadContextWords = processSeedLine ($seedLine2);
    close SEED;

    opendir $dh, "$dir/pos/"||die "Could not open the directory";
    my @reviews = readdir $dh;
    close $dh;
    my $total = 0;
    my $totCorrect = 0;
    my $totPos = 0;
    my $nuetral = 0 ;
    my $incorrect = 0;
    my @revWords;

	# make list of sentences from web containing seed words
	#
    getWebSentences($seedLine1,$seedLine2);

#    print "done\n" ;
    print "\nReview Name\tActual Sentiment\tSentiment We Tag\n\n";

    foreach $review (@reviews)
    {

	if (($review eq ".")||($review eq ".."))
	{
	    next;
	}
	# Return list of "relevant" unigrams from current review
	my @revWords = giveMeWordsIn("$dir/pos/$review"); 
	foreach( @revWords ) {
	    $revWords{$_} = 0 ;
	}
	
	my $revScore = 0;

	# Evaluate each unigram, and use the sum to assign semantic orientation.

	foreach my $revWord (keys %revWords)
	{
	    my $sizeG = 0;
	    my $good = evaluateBigram($sizeG,$revWord,%purgedGoodContextWords);
	    my $sizeB = 0;
	    my $bad = evaluateBigram($sizeB,$revWord,%purgedBadContextWords);

	    my $goodScore = 0; 
	    my $badScore = 0; 
	    if ($sizeG>0)
	    {
		$goodScore = $good/$sizeG;
	    }
	    if ($sizeB>0)
	    {
		$badScore = $bad/$sizeB;
	    }

	    my $revWordScore = $goodScore - $badScore;
	    $revScore += $revWordScore;
	}
	# assign orientation: positive, negative, nuetral.	    
	# 
	my $judgement = "Neutral";
	if ($revScore > 0.5)
	{
	    $judgement = "Positive";
	    $totCorrect++;
	} elsif ($revScore < (-0.5))
	{
	    $judgement = "Negative";
	    $incorrect++;
	}
	else{ $nuetral++ ; }

	print TRACE "$review\t\tActual type: Positive\tScore: $revScore\tOur Judgement: $judgement\n";
	$total++;
	print "$review\tPositive\t$judgement\n";
    }

#
# Repeat the above process for the files in the neg folder.
#

    opendir $dh, "$dir/neg";
    @reviews = ();
    @reviews = readdir $dh;
    close $dh;

    foreach $review (@reviews)
    {
	if (($review eq ".")||($review eq ".."))
	{
	    next;
	}
	
	my @revWords = giveMeWordsIn("$dir/neg/$review"); 
		foreach( @revWords ) {
			$revWords{$_} = 0 ;
	}	
	
	my $revScore = 0;
	
	foreach my $revWord (keys %revWords)
	{
	    my $sizeG = 0;
	    my $good = evaluateBigram($sizeG,$revWord,%purgedGoodContextWords);
	    my $sizeB = 0;
	    my $bad = evaluateBigram($sizeB,$revWord,%purgedBadContextWords);

	    my $goodScore = 0; 
	    my $badScore = 0; 
	    if ($sizeG>0)
	    {
		$goodScore = $good/$sizeG;
	    }
	    if ($sizeB>0)
	    {
		$badScore = $bad/$sizeB;
	    }

	    my $revWordScore = $goodScore - $badScore;
	    $revScore += $revWordScore;
	}
	    
	my $judgement = "Neutral";

	if ($revScore > 0.5)
	{
	    $judgement = "Positive";
	    $incorrect++;
	} elsif ($revScore < (-0.5))
	{
	    $judgement = "Negative";
	    $totCorrect ++ ;
	}
	else{ $nuetral++ ; }

	print TRACE "$review\t\tActual type: Negative\tScore: $revScore\tOur Judgement: $judgement\n";
	print "$review\tNegative\t$judgement\n";	
	$total++;
    }
	
    print TRACE "SUMMARY\n***********************************\n";
    print TRACE "\tTotal Reviews = $total\n";
    print TRACE "\tCorrect = $totCorrect\n";
    print "$totCorrect $incorrect $nuetral\n";

    if ($total == 0)
    {
	print TRACE "\tPRECISION = 0\n\tRECALL = 0\n";
	close TRACE;
	return 1;
    }  
    
    my $precision = $totCorrect/($total-$nuetral);
    my $recall = $totCorrect/$total;

    $_[1] = $precision;
    $_[2] = $recall;
    
	# Calculate precission and recall and write to data file.
    print TRACE "\tPRECISION = $precision\n\tRECALL = $recall\n";
    close TRACE;


	#print "Number of nuetral responses: $nuetral\n" ;
	
	#Return fractional portion correct.
	#
    return ($totCorrect/$total);
}

sub giveMeWordsInOld
{
    my $review = $_[0];
    my %outHash = ();
    my $fh = FILEinGiveMeWordsIn;
    if (!(open $fh, "<$review"))
    {
	print "Open failed on $review\n";
	return %outHash;
    }

    while(my $line = <$fh>)
    {
	my @words = split(/([\s\n\t,.!\?])+/, $line);
	foreach my $word (@words)
	{
	    if ($word !~ /^\w+$/)
	    {
		#print "$word\n";
		next;
	    }
	    my %pos = ();
	    %pos = ldoc::getPOSFor($word);
	    if (exists($pos{"adj"}))
	    {
		$outHash{$word} = 0;
	    }
	    if (exists($pos{"adv"}))
	    {
		$outHash{$word} = 0;
	    }
	}
    }
    close $fh;
    return %outHash;
}
	    

sub arrayToHash
{
    my @inputArray = @_;
    my %outHash = ();
    while (@inputArray)
    {
	my $pos = shift(@inputArray);
	$outHash{$pos} = shift (@inputArray);
    }
    return %outHash;
}

sub purge
{
    my %outHash = ();
    foreach my $defWord (@_)
    {
	$outHash{$defWord}=0;
    }

    foreach my $key (keys %outHash)
    {
	if ($key !~ /^\w+$/)
	{
	    delete($outHash{$key});
	    next;
	}
	my @tempArr = ldoc::Lookup($key);
	shift (@tempArr);
	my %def = arrayToHash(@tempArr);

	my $deleteKey = 1;
	foreach my $key1 (keys %def)
	{
	    if ($key1 eq "adj")
	    {
		$deleteKey = 0;
		last;
	    }

	    if ($key1 eq "adv")
	    {
		$deleteKey = 0;
		last;
	    }
	}

	if ($deleteKey)
	{
	    delete ($outHash{$key});
	}
    }
    return %outHash;
}


#This function computes score for a word (not bigram. 
#So in that the name is a misnomer :-) ) in the review. 
#It makes use of good and bad contexts for this score computation. 

sub evaluateBigram
{
    my @tempIpArray = @_;
    shift (@tempIpArray);
    shift (@tempIpArray);
    my %purgedContextWords = @tempIpArray;

    my @words = ();
    @words = split (/\s+/,$_[1]);
    my $biGram = $_[1];
#    my $context = $_[2];

#    open FILE2, ">Allinone";    
#    open FILE1, ">123";
#    system ("chmod 755 123 Allinone");
    
    my $defs = "";
    foreach my $word(@words)
    {
	my @tempArr = ldoc::Lookup($word);
	shift (@tempArr); #Throw the headword out.
	my %def = arrayToHash(@tempArr);
	foreach my $pos (keys %def)
	{
	    if (($pos eq "adj")||($pos eq "adv"))
	    {
		foreach my $key (keys %{$def{$pos}})
		{
		    $defs .= " $key";
		}
	    }
	}
    }

    my @defWords = split (/([\s\n\t,.!\?])+/, $defs);    
    my %purgedDefWords = purge (@defWords);
    $purgedDefWords{$biGram}=0;


    my $matches = 0; 
    my $size = 0;
    foreach my $defWord (keys %purgedDefWords)
    {
	if (exists ($purgedContextWords{$defWord}))
	{
	    $matches++;
	}
	$size++;
    }

    $_[0] = $size;
    return $matches;
}	   

sub filterStr
{
    my $context = shift(@_);
    if (!(defined($context)))
    {
	return "";
    }
    my $tempStr = shift(@_);
    if (!(defined($tempStr)))
    {
	return $context;
    }
    
    my @stopList = split (/\s*,\s*/,$tempStr);


    foreach my $stopWord(@stopList)
    {
	while ($context =~ /\b$stopWord\b/)
	{
	    $context = $`.$';
	}
    }
    return $context;
}

sub generateListnFile
{
    my %outList = ();
    my $ipStr = shift (@_);

    if (!(defined($ipStr)))
    {
	return %outList;
    }

    if (!($ipStr =~ /:/))
    {
	return %outList;
    }

    my $word = $`;
    my $tempStr = $';
    $word =~ s/\s*$//;
    $tempStr =~ s/^\s*//;

    my $wordListFile = $word."List";
    my $fh = FILEinGenerateListnFile;
    my $generateListFile = 0;

    if (!(open $fh,"<$wordListFile"))
    {
	$generateListFile = 1;
    }
    else
    {
	my $firstLine = <$fh>;
	chomp($firstLine);
	
	if ($firstLine ne $tempStr)
	{
	    $generateListFile = 1;
	    close $fh;
	}	
    }

    if ($generateListFile)
    {
	my @words = split(/,\s*/,$tempStr);
	foreach my $wrd (@words)
	{
	    my %list = makeWordList($wrd);
	    foreach my $key (keys %list)
	    {
		$outList{$key} = 1;
	    }
	}

 
	if (!(open $fh,">$wordListFile"))
	{
	    return %outList;
	}
	
	print $fh "$tempStr\n";
  
	foreach my $key (keys %outList)
	{
	    print $fh "$key\n";
	}
	close $fh;	
	system "chmod 755 $wordListFile";
    }
    else
    {
	while (<$fh>)
	{
	    chomp();
	    $outList{$_} = 0;
	}
    }

    return %outList;
}    

sub createContext
{
    $outStr = "";
    if (!(exists($_[0])))
    {
	return $outStr;
    }

    my %globalList = generateListnFile($_[0]);

    foreach my $key (keys %globalList)
    {
	my %hash = ();
	my @out = ldoc::Lookup($key);
	shift (@out); #Throw the headword out
	
	while (@out)
	{
	    my $pos = shift(@out);
	    if (($pos eq "adj")||($pos eq "adv")||($pos eq "adj, adv"))
	    {
		$hash{$pos} = shift (@out);
	    }
	    else
	    {
		shift (@out);
	    }
	}
	
	foreach my $key (keys %hash)
	{
	    foreach my $def (keys (%{$hash{$key}}))
		{
		    $outStr .="\n$def";
		}
	}
    }
    return $outStr;
}

##This function return for a given word, the list of words 
##that share both a semantic code and a thesaurus entry with it. 

sub makeWordList
{
	my $word = shift (@_);

	my %tempList = ldoc::getWordsWithSameSem($word);

	my %bigMacList = QueryMac::thesFindList($word); 

	my %list = ();
	foreach my $key (keys %tempList)
	{
		foreach my $val1 (keys %{$tempList{$key}})
		{	
		    foreach my $val2 (keys %bigMacList)
		    {
			if ($val1 eq $val2)
			{
			    $list{$val1}=0;
			}
		    }
		}
	}
	return (%list);	
}

##This function retrieves all the sentences from web that contain
##the given word or a word that has same semantic code and
##same thesaurus entry as the given word.
  
sub makeSentenceList
{
    my @wordList = @_ ;
    my %wordList = () ;
    my @sentences = ();
    #my $word = shift (@_);

    if ( scalar @wordList == 0 )
    {
	return @sentences;
    }

    foreach ( @wordList ) 
    {
	# expand list of relevant words
	my %tempHash = makeWordList($_);
	foreach my $key (keys %tempHash)
	{
	    $wordList{$key} = 0 ;
	}
    }

    #my %wordList = makeWordList($word);
    #$wordList{$seedWord} = 0 ;
    my $index = 0;
    foreach my $key (sort keys %wordList)
    {
	my @tempStore = webSearch::master($key);
	# concatenate each result to list of sentences
	foreach my $sntc (@tempStore)
	{
	    $sentences[$index++] = $sntc;
	}
    }
	
#    foreach (@sentences)
#    {
#	print "************\n$_\n**************\n";
#    }
#    print "Total = ",scalar @sentences;
    return (@sentences);
}

sub phraseCollector {

    my $file = shift ;
    my $phraseStr ;
    my @unigrams ;
    my @bigrams ;
    my $first ;
    my $second ;
    my $FH = FileInPhraseCollector ;
    $phraseCollectorFlag = 0 ; # global flag

    open $FH,  "$file" || die "Couldn't open example for reading: $!\n" ;

    if( $phraseCollectorFlag == 0 ) { QueryMac::initialize() ; $phraseCollectorFlag = 1 ; }
    
	
    # Get the single line
    $phraseStr = <$FH> ;
    #print "**$phraseStr**" ;
    # input file is a single line, so this should
    # collect all the bigrams
    #
    # Pull out individual words
    @unigrams = ( $phraseStr =~ /(\b\w+?)\W+/g ) ;
    
    #Concatenate into bigrams
    #
    foreach $word (@unigrams ) {
	$second = $word ;
	if (!defined($first)) {$first = $word ; next ;}
	else { 
	    # if first word is an adjective or adverb add to final list
	    # otherwise , check second word.
	    if( QueryMac::lookup($first, "adjective") || QueryMac::lookup($first, "adverb") ) { 
		$phrase = $first." ".$second ;
		$phrase = uc($phrase) ;
		push @bigrams, $phrase ;
		$first = $second ;
	    }
	    else {
		# if it is not, then skip
		if(QueryMac::lookup($second, "adjective") || QueryMac::lookup($second, "adverb") ) {
		    $phrase = $first." ".$second ;
		    $phrase = uc($phrase) ;
		    push @bigrams, $phrase ;
		    $first = $second ;
		}
		else { next ; }
	    }
	}
    }
    #print "returning bigrams\n" ;
    #foreach (@bigrams) { print ; }
    return @bigrams ;

}
